#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <limits>
#include <algorithm>
using namespace std;

#include "Raytrace.h"
#include "SimpleString.h"
#include "Texture.h"
#include "Perlin.h"
#include "Scene.h"

bool hitSphere(const ray &r, const sphere& s, float &t)
{
	vecteur dist = s.pos - r.start;
	float B = (r.dir.x * dist.x + r.dir.y * dist.y + r.dir.z * dist.z);
	float D = B*B - dist*dist + s.size * s.size;
	if (D < 0.0f) return false;
	float t0 = B - sqrtf(D);
	float t1 = B + sqrtf(D);
	bool retvalue = false;
	if ((t0 > 0.1f ) && (t0 < t))
	{
		t = t0;
		retvalue = true;
	}
	if ((t1 > 0.1f ) && (t1 < t))
	{
		t = t1;
		retvalue = true;
	}
	return retvalue;
}

color addRay(ray viewRay, scene &myScene, context myContext)
{
	color output = {0.0f, 0.0f, 0.0f}; 
	float coef = 1.0f;
	int level = 0;
	do 
	{
        point ptHitPoint;
        vecteur vNormal;
        material currentMat;
        {
		    int currentBlob=-1;
		    int currentSphere=-1;
		    float t = 20000.0f;
		    for (unsigned int i = 0; i < myScene.blbTab.size() ; ++i)
		    {
			    if (isBlobIntersected(viewRay, myScene.blbTab[i], t)) {
				    currentBlob = i;
			    }
		    }
		    for (unsigned int i = 0; i < myScene.sphTab.size() ; ++i)
		    {
			    if (hitSphere(viewRay, myScene.sphTab[i], t))
			    {
				    currentSphere = i;
                    currentBlob = -1;
			    }
		    }
            if (currentBlob != -1)
            {
                ptHitPoint  = viewRay.start + t * viewRay.dir;
                blobInterpolation(ptHitPoint, myScene.blbTab[currentBlob], &myScene.matTab[0], vNormal, currentMat);
                float temp = vNormal * vNormal;
		        if (temp == 0.0f)
                    break;
		        vNormal = invsqrtf(temp) * vNormal;
		        currentMat = myScene.matTab[myScene.blbTab[currentBlob].materialId];
            }
            else if (currentSphere != -1)
            {
                ptHitPoint  = viewRay.start + t * viewRay.dir;
		        vNormal = ptHitPoint - myScene.sphTab[currentSphere].pos;
		        float temp = vNormal * vNormal;
		        if (temp == 0.0f)
			        break;
		        temp = invsqrtf(temp);
		        vNormal = temp * vNormal;
		        currentMat = myScene.matTab[myScene.sphTab[currentSphere].materialId];
            }
            else
            {
                break;
            }
        }

        float bInside;

        if (vNormal * viewRay.dir > 0.0f)
        {
            vNormal = -1.0f * vNormal;
            bInside = true;
        }
        else
        {
            bInside = false;
        }

        if (currentMat.bump)
        {
            float noiseCoefx = float(noise(0.1 * double(ptHitPoint.x), 0.1 * double(ptHitPoint.y),0.1 * double(ptHitPoint.z)));
            float noiseCoefy = float(noise(0.1 * double(ptHitPoint.y), 0.1 * double(ptHitPoint.z),0.1 * double(ptHitPoint.x)));
            float noiseCoefz = float(noise(0.1 * double(ptHitPoint.z), 0.1 * double(ptHitPoint.x),0.1 * double(ptHitPoint.y)));
            
            vNormal.x = (1.0f - currentMat.bump ) * vNormal.x + currentMat.bump * noiseCoefx;  
            vNormal.y = (1.0f - currentMat.bump ) * vNormal.y + currentMat.bump * noiseCoefy;  
            vNormal.z = (1.0f - currentMat.bump ) * vNormal.z + currentMat.bump * noiseCoefz;  
            
            float temp = vNormal * vNormal;
            if (temp == 0.0f)
            break;
            temp = invsqrtf(temp);
            vNormal = temp * vNormal;
        }
        
        float fViewProjection = viewRay.dir * vNormal;
        float fReflectance, fTransmittance;
        float fCosThetaI, fSinThetaI, fCosThetaT, fSinThetaT;

        if(((currentMat.reflection != 0.0f) || (currentMat.refraction != 0.0f) ) && (currentMat.density != 0.0f))
        {
            // material de type verre on va calculer le coefficient de reflexion de fresnel

            float fDensity1 = myContext.fRefractionCoef; 
            float fDensity2;
            if (bInside)
            {
                // on considere que le rayon incident traverse un ether proche du vide (ou de l'air)
                // en thorie il faudrait d'abord calculer pour savoir si l'extrieur
                // se trouve dans un autre objet mais a dpasse le cadre de notre code.
                fDensity2 = context::getDefaultAir().fRefractionCoef;
            }
            else
            {
                fDensity2 = currentMat.density;
            }

            // ici on tient compte du fait que le dplacement de la lumiere
            // est symetrique, de l'observateur a la source de la lumiere ou de la source de lumiere
            // a l'observateur. On fait donc les calculs du coefficient en prenant en compte le rayon provenant 
            // du point d'observation.
            fCosThetaI = fabsf(fViewProjection); 

            if (fCosThetaI >= 0.999f) 
            {
                // cas du rayon incident selon la normale a la surface
                fReflectance = (fDensity1 - fDensity2) / (fDensity1 + fDensity2);
                fReflectance = fReflectance * fReflectance;
                fSinThetaI = 0.0f;
                fSinThetaT = 0.0f;
                fCosThetaT = 1.0f;
            }
            else 
            {
                fSinThetaI = sqrtf(1 - fCosThetaI * fCosThetaI);
                // le signe de Sin ThetaI n'a aucune importance, c'est le meme que celui de Sin ThetaT
                // et ils s'annihilent mutuellement dans le calcul des coefficients de Reflexion
                fSinThetaT = (fDensity1 / fDensity2) * fSinThetaI;
                if (fSinThetaT * fSinThetaT > 0.9999f)
                {
                    // On a atteint l'angle critique au dela duquel
                    // la surface est entierement reflexive
                    fReflectance = 1.0f ;
                    fCosThetaT = 0.0f;
                }
                else
                {
                    fCosThetaT = sqrtf(1 - fSinThetaT * fSinThetaT);
                    // tout d'abord la reflectance de la lumiere polaris dans le 
                    // plan orthogonal au plan de reflexion
                    float fReflectanceOrtho = (fDensity2 * fCosThetaT - fDensity1 * fCosThetaI ) 
                        / (fDensity2 * fCosThetaT + fDensity1  * fCosThetaI);
                    fReflectanceOrtho = fReflectanceOrtho * fReflectanceOrtho;
                    // Ensuite la reflectance de la lumiere polaris dans le 
                    // plan parallele au plan de reflexion
                    float fReflectanceParal = (fDensity1 * fCosThetaT - fDensity2 * fCosThetaI )
                        / (fDensity1 * fCosThetaT + fDensity2 * fCosThetaI);
                    fReflectanceParal = fReflectanceParal * fReflectanceParal;

                    // le coefficient de reflectance global est gal  la moyenne
                    // des deux pour une lumire de base non polarise
                    fReflectance =  0.5f * (fReflectanceOrtho + fReflectanceParal);
                }
            }
        }
        else
        {
            // reflexion de type "metal", la reflectance est la meme dans toutes les directions
            // Les mtaux sont conducteurs et induisent un changement dans la polarisation de la lumire
            // que l'on ignore royalement.
            fReflectance = 1.0f;
            fCosThetaI = 1.0f;
            fCosThetaT = 1.0f;
        }

        fTransmittance = currentMat.refraction * (1.0f - fReflectance);
        fReflectance = currentMat.reflection * fReflectance;

        float fTotalWeight = fReflectance + fTransmittance;
        bool bDiffuse = false;

        if (fTotalWeight > 0.0f)
        {
            float fRoulette = (1.0f / RAND_MAX) * rand();
        
            if (fRoulette <= fReflectance)
            {
                coef *= currentMat.reflection;

                // Le rayon viewRay est rflechi par la normale.
                // Pour calculer le rayon reflchi on fait la projection
                // du vecteur direction du rayon vue
                // sur la direction de la normale.
                // Pour avoir le vecteur tangent il faudrait retrancher
                // cette projection au vecteur direction
                // mais ici on veut la reflexion et donc il faut retrancher
                // la projection deux fois au vecteur direction.

                float fReflection = - 2.0f * fViewProjection;

                // on fait partir le nouveau rayon du point d'intersection avec la sphre courante
                // et on le dirige selon le vecteur reflexion que l'on vient de calculer
                viewRay.start = ptHitPoint;
		        viewRay.dir += fReflection * vNormal;
            }
            else if(fRoulette <= fTotalWeight)
            {
                coef *= currentMat.refraction;
                float fOldRefractionCoef = myContext.fRefractionCoef;
                if (bInside) 
                {
                    myContext.fRefractionCoef = context::getDefaultAir().fRefractionCoef;
                }
                else
                {
                    myContext.fRefractionCoef = currentMat.density;
                }

                // ici on calcule le rayon transmis avec la formule de Snell-Descartes
                viewRay.start = ptHitPoint;

                viewRay.dir = viewRay.dir + fCosThetaI * vNormal;
                viewRay.dir = (fOldRefractionCoef / myContext.fRefractionCoef) * viewRay.dir;
                viewRay.dir += (-fCosThetaT) * vNormal;
            }
            else
            {
                bDiffuse = true;
            }
        }
        else
        {
            bDiffuse = true;
        }

        // Maintenant qu'on a dtermin quelle tait le point visible
        // sur l'cran et dans quel direction pointait la surface
        // en ce point, il reste  calculer la valeur d'clairement
        // L encore il faut parcourir la liste des lumires
        // et dterminer celles qui ont une influence sur notre objet.

        if (!bInside && bDiffuse)
        {

            ray lightRay;
            lightRay.start = ptHitPoint;
            for (unsigned int j = 0; j < myScene.lgtTab.size() ; ++j)
            {
                light currentLight = myScene.lgtTab[j];

                lightRay.dir = currentLight.pos - ptHitPoint;
                float fLightProjection = lightRay.dir * vNormal;

                if ( fLightProjection <= 0.0f )
                    continue;

                float lightDist = lightRay.dir * lightRay.dir;
                {
                    float temp = lightDist;
			        if ( temp == 0.0f )
				        continue;
                    temp = invsqrtf(temp);
			        lightRay.dir = temp * lightRay.dir;
                    fLightProjection = temp * fLightProjection;
                }

                bool inShadow = false;
                {
                    float t = lightDist;
                    for (unsigned int i = 0; i < myScene.sphTab.size() ; ++i)
                    {
                        if (hitSphere(lightRay, myScene.sphTab[i], t))
                        {
                            inShadow = true;
                            break;
                        }
                    }
                    for (unsigned int i = 0; i < myScene.blbTab.size() ; ++i)
                    {
                        if (isBlobIntersected(lightRay, myScene.blbTab[i], t)) {
                            inShadow = true;
                            break;
                        }
                    }
                }

                if (!inShadow && (fLightProjection > 0.0f))
                {

                    float lambert = (lightRay.dir * vNormal) * coef;
                    float noiseCoef = 0.0f;
                    switch(currentMat.type)
                    {
                    case material::turbulence:
                        {
                            for (int level = 1; level < 10; level ++)
                            {
                                noiseCoef += (1.0f / level )  
                                    * fabsf(float(noise(level * 0.05 * ptHitPoint.x,  
                                                        level * 0.05 * ptHitPoint.y,
                                                        level * 0.05 * ptHitPoint.z)));
                            };
                            output = output +  coef * (lambert * currentLight.intensity)  
                                    * (noiseCoef * currentMat.diffuse + (1.0f - noiseCoef) * currentMat.diffuse2);
                        }
                        break;
                    case material::marble:
                        {
                        for (int level = 1; level < 10; level ++)
                        {
                            noiseCoef +=  (1.0f / level)  
                            * fabsf(float(noise(level * 0.05 * ptHitPoint.x,  
                                                level * 0.05 * ptHitPoint.y,  
                                                level * 0.05 * ptHitPoint.z)));
                        };
                        noiseCoef = 0.5f * sinf( (ptHitPoint.x + ptHitPoint.y) * 0.05f + noiseCoef) + 0.5f;
                        output = output +  coef * (lambert * currentLight.intensity)  
                            * (noiseCoef * currentMat.diffuse + (1.0f - noiseCoef) * currentMat.diffuse2);
                        }
                        break;
                    default:
                        {
                        output.red += lambert * currentLight.intensity.red * currentMat.diffuse.red;
                        output.green += lambert * currentLight.intensity.green * currentMat.diffuse.green;
                        output.blue += lambert * currentLight.intensity.blue * currentMat.diffuse.blue;
                        }
                        break;
                    }

                    // Blinn 
                    // La direction de Blinn est exactement  mi chemin entre le rayon
                    // lumineux et le rayon de vue. 
                    // On calcule le vecteur de Blinn et on le rend unitaire
                    // puis on calcule le cofficient de blinn
                    // qui est la contribution spculaire de la lumire courante.

                    vecteur blinnDir = lightRay.dir - viewRay.dir;
                    float temp = blinnDir * blinnDir;
                    if (temp != 0.0f )
                    {
                        float blinn = invsqrtf(temp) * max(fLightProjection - fViewProjection , 0.0f);
                        blinn = coef * powf(blinn, currentMat.power);
                        output += blinn *currentMat.specular  * currentLight.intensity;
                    }
                }
            }
            coef = 0.0f ;
        }

        level++;
    } while ((coef > 0.0f) && (level < 10));  

    if (coef > 0.0f)
    {
        output += coef * readCubemap(myScene.cm, viewRay);
    }
    return output;
}

float AutoExposure(scene &myScene)
{
    #define ACCUMULATION_SIZE 256
	float exposure = -1.0f;
	float accufacteur = float(max(myScene.sizex, myScene.sizey));
    float projectionDistance = 0.0f;

    if ((myScene.persp.type == perspective::conic) && myScene.persp.FOV > 0.0f && myScene.persp.FOV < 189.0f)
    {
        // On triangule avec l'angle du FOV divis par deux et 
        // la taille du viewport divis par deux
        projectionDistance = 0.5f * myScene.sizex / tanf (float(PIOVER180) * 0.5f * myScene.persp.FOV);
    }

	accufacteur = accufacteur / ACCUMULATION_SIZE;
	color mediumPoint = {0.0f, 0.0f, 0.0f};
	const float mediumPointWeight = 1.0f / (ACCUMULATION_SIZE*ACCUMULATION_SIZE);
	for (int y = 0; y < ACCUMULATION_SIZE; ++y) {
		for (int x = 0 ; x < ACCUMULATION_SIZE; ++x) {

            if (myScene.persp.type == perspective::orthogonal || projectionDistance == 0.0f)
            {
                ray viewRay = { {float(x)*accufacteur, float(y) * accufacteur, -10000.0f}, { 0.0f, 0.0f, 1.0f}};
			    color currentColor = addRay (viewRay, myScene, context::getDefaultAir());
			    mediumPoint = mediumPoint + mediumPointWeight * (currentColor * currentColor);
            }
            else
            {
                vecteur dir = {(float(x)*accufacteur - 0.5f * myScene.sizex) / projectionDistance, 
                                (float(y) * accufacteur - 0.5f * myScene.sizey) / projectionDistance, 
                                1.0f}; 

                float norm = dir * dir;
                // je ne pense pas que a puisse arriver mais on n'est jamais trop prudent
		        if (norm == 0.0f) 
			        break;
                dir = invsqrtf(norm) * dir;

                ray viewRay = { {0.5f * myScene.sizex,  0.5f * myScene.sizey, 0.0f}, {dir.x, dir.y, dir.z} };
			    color currentColor = addRay (viewRay, myScene, context::getDefaultAir());
			    mediumPoint = mediumPoint + mediumPointWeight * (currentColor * currentColor);
            }
		}
	}
	
    // Code d'autoexposition, on place le carr moyen  une valeur de gris intermdiaire

	float mediumLuminance = sqrtf(0.2126f * mediumPoint.red + 0.715160f * mediumPoint.green + 0.072169f * mediumPoint.blue);
	if (mediumLuminance > 0.001f)
	{
		exposure = logf(0.6f) / mediumLuminance;
	}

    return exposure;
}

bool draw(char* outputName, scene &myScene)
{
	int x, y;
	ofstream imageFile(outputName,ios_base::binary);
	if (!imageFile)
		return false;
	// specific TGA handling code
	imageFile.put(0).put(0);
	imageFile.put(2);                  /* uncompressed RGB */

	imageFile.put(0).put(0);
	imageFile.put(0).put(0);
	imageFile.put(0);

	imageFile.put(0).put(0);           /* X origin */
	imageFile.put(0).put(0);           /* y origin */

	imageFile.put((myScene.sizex & 0x00FF)).put((myScene.sizex & 0xFF00) / 256);
	imageFile.put((myScene.sizey & 0x00FF)).put((myScene.sizey & 0xFF00) / 256);
	imageFile.put(24);                 /* 24 bit bitmap */
	imageFile.put(0);
	
    float exposure = AutoExposure(myScene);
    // on fait des maths ici parce que nos paramtres sont
    // implicites, on donne le FOV horizontal (le champ de vision)
    // en paramtre et il faut en dduire  quelle distance 
    // virtuelle se trouve notre plan de projection du point d'origine. 
    float projectionDistance = 0.0f;

    if ((myScene.persp.type == perspective::conic) && myScene.persp.FOV > 0.0f && myScene.persp.FOV < 189.0f)
    {
        // On triangule avec l'angle du FOV divis par deux et 
        // la taille du viewport divis par deux
        projectionDistance = 0.5f * myScene.sizex / tanf (float(PIOVER180) * 0.5f * myScene.persp.FOV);
    }

	for (y = 0; y < myScene.sizey; ++y)
	{
		for (x = 0 ; x < myScene.sizex; ++x)
		{
			color output = {0.0f, 0.0f, 0.0f};
			for (float fragmentx = float(x) ; fragmentx < x + 1.0f; fragmentx += 0.5f )
			for (float fragmenty = float(y) ; fragmenty < y + 1.0f; fragmenty += 0.5f )
			{
				float sampleRatio = 0.25f;
                color temp = {0.0f, 0.0f, 0.0f};
                float fTotalWeight = 0.0f;

                if (myScene.persp.type == perspective::orthogonal || projectionDistance == 0.0f)
                {
                    ray viewRay = { {fragmentx, fragmenty, -10000.0f}, { 0.0f, 0.0f, 1.0f}};
                    for (int i = 0; i < myScene.complexity; ++i)
                    {                  
                        color rayResult = addRay (viewRay, myScene, context::getDefaultAir());
                        fTotalWeight += 1.0f; 
                        temp += rayResult;
                    }
                    temp = (1.0f / fTotalWeight) * temp;
                }
                else
                {
                    vecteur dir = {(fragmentx - 0.5f * myScene.sizex) / projectionDistance, 
                                (fragmenty - 0.5f * myScene.sizey) / projectionDistance, 
                                1.0f}; 

                    float norm = dir * dir;
                    // je ne pense pas que a puisse arriver mais on n'est jamais trop prudent
                    if (norm == 0.0f) 
	                    break;
                    dir = invsqrtf(norm) * dir;
                    // le point de dpart est toujours le centre optique de la camra
                    // il peut tre perturb plus tard pour un effet de profondeur de champ
                    point start = {0.5f * myScene.sizex,  0.5f * myScene.sizey, 0.0f};
                    // le point vis est le point invariant du pixel courant,
                    // c'est  dire que par construction tous les rayons qui participent
                    // au pixel courant doivent passer par ce point
                    point ptVise = start + myScene.persp.clearPoint * dir;

                    for (int i = 0; i < myScene.complexity; ++i)
                    {                  
                        ray viewRay = { {start.x, start.y, start.z}, {dir.x, dir.y, dir.z} };

                        if (myScene.persp.dispersion != 0.0f)
                        {
                            vecteur vDisturbance;                        
                            vDisturbance.x = (myScene.persp.dispersion / RAND_MAX) * (1.0f * rand());
                            vDisturbance.y = (myScene.persp.dispersion / RAND_MAX) * (1.0f * rand());
                            vDisturbance.z = 0.0f;

                            viewRay.start = viewRay.start + vDisturbance;
                            viewRay.dir = ptVise - viewRay.start;
                            
                            norm = viewRay.dir * viewRay.dir;
                            if (norm == 0.0f)
                                break;
                            viewRay.dir = invsqrtf(norm) * viewRay.dir;
                        }
                        color rayResult = addRay (viewRay, myScene, context::getDefaultAir());
                        fTotalWeight += 1.0f;
                        temp += rayResult;
                    }
                    temp = (1.0f / fTotalWeight) * temp;
                }

                // exposition photo
	            temp.blue = (1.0f - expf(temp.blue * exposure));
	            temp.red =  (1.0f - expf(temp.red * exposure));
	            temp.green = (1.0f - expf(temp.green * exposure));
                
	            output += sampleRatio * temp;
			}
						
			// gamma correction
            
			float invgamma = 0.45f; // c'est la valeur fixe par le standard sRGB = 1 / 2.2
			output.blue  = powf(output.blue, invgamma);
			output.red   = powf(output.red, invgamma);
			output.green = powf(output.green, invgamma);
            
            if (y < 10)
            {
                if ((x / 10) & 1)
                {
                    imageFile.put((unsigned char)186).put((unsigned char)186).put((unsigned char)186);
                }
                else if ( y & 1)
                {
                    imageFile.put((unsigned char)255).put((unsigned char)255).put((unsigned char)255);
                }
                else
                {
                    imageFile.put((unsigned char)0).put((unsigned char)0).put((unsigned char)0);
                }
            }
            else
            {
                imageFile.put((unsigned char)min(output.blue*255.0f,255.0f)).put((unsigned char)min(output.green*255.0f, 255.0f)).put((unsigned char)min(output.red*255.0f, 255.0f));
            }
        }
	}
	return true;
}

int main(int argc, char* argv[])
{
	if (argc < 3)
    {
        cout << "Usage : Raytrace.exe Scene.txt Output.tga" << endl;
		return -1;
    }
	scene myScene;
	if (!init(argv[1], myScene))
    {
        cout << "Echec  l'ouverture du fichier scene." << endl;
		return -1;
    }
	if (!draw(argv[2], myScene))
    {
        cout << "Echec  la cration du fichier image." << endl;
		return -1;
    }
	return 0;
}
